@@ -12,10 +12,17 @@ use base qw(Net::DNS::Resolver::Base DynaLoader), OS_CONF;
 our $VERSION;
 
 BEGIN {
-	$VERSION = '1.24_02';
+	$VERSION = '1.25_01';
 	eval { __PACKAGE__->bootstrap($VERSION) };
 }
 
+use constant UNBOUND_VERSION => eval {
+        my $version = Net::DNS::Resolver::libunbound->VERSION();
+        my ($v1,$v2,$v3) = split /\D/, $version;
+        ( $v1 * 100 + $v2 ) * 100 + $v3;
+};
+
+
 use constant UB_CONTEXT => 'Net::DNS::Resolver::Unbound::Context';
 
 
@@ -90,7 +97,7 @@ sub new {
 	my ( $class, @args ) = @_;
 	my $self = $class->SUPER::new();
 	$self->nameservers( $self->SUPER::nameservers );
-	$self->_finalise_config;				# default configuration
+	$self->_finalise_config;				# base configuration
 	$self->{update} = {} if @args;				# force context rebuild
 	while ( my $attr = shift @args ) {
 		my $value = shift @args;
@@ -126,9 +133,11 @@ By default, DNS queries are sent to the IP addresses listed in
 sub nameservers {
 	my ( $self, @nameservers ) = @_;
 	if ( defined wantarray ) {
-		$self->_finalise_config;
-		my %config = %{$self->{config}};
-		return @{$config{set_fwd}};
+		my $config   = $self->{config};			# merge config updates
+		my $update   = $self->{update};
+		my ($setfwd) = grep {$_} $update->{set_fwd}, $config->{set_fwd};
+		my @value    = map  { ref($_) ? @$_ : $_ } $setfwd;
+		return @value;
 	}
 	$self->set_fwd() unless @nameservers;
 	$self->set_fwd($_) foreach @nameservers;
@@ -241,7 +250,7 @@ as backup servers.
 
 sub set_fwd {
 	my ( $self, @fwd ) = @_;
-	return $self->_config( 'set_fwd', [@fwd] );
+	return $self->_config( 'set_fwd', @fwd );
 }
 
 
@@ -276,7 +285,7 @@ use constant SET_STUB => UB_CONTEXT->can('set_stub');
 
 sub set_stub {
 	my ( $self, $zone, $address, $prime ) = @_;
-	return SET_STUB ? $self->_config( 'set_stub', [[$zone, $address, $prime]] ) : undef;
+	return SET_STUB ? $self->_config( 'set_stub', $zone, $address, $prime ) : undef;
 }
 
 
@@ -437,19 +446,26 @@ sub string {
 	my $self = shift;
 	$self = $self->new() unless ref($self);
 
-	my $image = <<END;
+	my ($force)  = ( grep( { $self->{$_} } qw(force_v6 force_v4) ),	  'force_v4' );
+	my ($prefer) = ( grep( { $self->{$_} } qw(prefer_v6 prefer_v4) ), 'prefer_v4' );
+	my $image    = <<END;
 ;; RESOLVER state:
-;; debug	$self->{debug}	ndots	$self->{ndots}
-;; defnames	$self->{defnames}	dnsrch	$self->{dnsrch}
 ;; searchlist	@{$self->{searchlist}}
+;; defnames	$self->{defnames}	dnsrch	$self->{dnsrch}
+;; ${prefer}	$self->{$prefer}	ndots	$self->{ndots}
+;; ${force}	$self->{$force}	debug	$self->{debug}
 END
+	$self->{update} ||= {};					# force config rebuild
 	$self->_finalise_config;
 	my %config = %{$self->{config}};
-	my $optref = $config{set_option} || [];		# sort option list
-	my %option = map { @$_ } @$optref;
+	my $optref = $config{set_option} || [];			# extract option hash
+	my %option = map {@$_} @$optref;
 	my @option;
-	foreach my $opt ( sort keys %option ) {
-		push @option, [$opt, $option{$opt}];
+
+	foreach my $opt ( sort keys %option ) {			# build sorted list
+		my $value = $option{$opt};
+		my @value = map { ref($_) ? @$_ : $_ } $value;
+		push @option, [$opt, $_] foreach @value;
 	}
 
 	my $format = ";; %s\t%s\n";
@@ -489,18 +505,16 @@ sub _decode_result {
 
 
 sub _config {
-	my ( $self, $name, $arg ) = @_;
-	$self->{ub_ctx} = Net::DNS::Resolver::Unbound::Context->new() unless $self->{update};
-	my $update = $self->{update} ||= {};			# collect context changes
-	if ( ref $arg ) {
-		my @arg = map { ref($_) ? @$_ : $_ } @$arg;
-		$self->{ub_ctx}->$name(@arg) if @arg;		# error check only
-		my $list = $update->{$name} ||= [];
-		push @$list, @$arg if @arg;			# append non-empty
-		$update->{$name} = [] unless @arg;		# explicit empty list
-	} elsif ( defined $arg ) {
-		$self->{ub_ctx}->$name($arg);			# error check only
-		$update->{$name} = $arg;			# [re]define scalar
+	my ( $self, $name, @arg ) = @_;
+	my $entry = ( scalar(@arg) == 1 ) ? $arg[0] : [@arg];
+	$self->{test_ctx} = Net::DNS::Resolver::Unbound::Context->new() unless $self->{update};
+	$self->{test_ctx}->$name(@arg) if @arg;			# error check only
+	my $state = $self->{update}->{$name};
+	if ( defined $state ) {					# second and subsequent entries
+		$state = $self->{update}->{$name} = [$state] unless ref $state;
+		push @$state, $entry;
+	} else {						# initial entry
+		$self->{update}->{$name} = ( scalar(@arg) > 1 ) ? [$entry] : $entry;
 	}
 	return;
 }
@@ -508,21 +522,29 @@ sub _config {
 
 sub _option {
 	my ( $self, $name, @arg ) = @_;
-	$self->{ub_ctx} = Net::DNS::Resolver::Unbound::Context->new() unless $self->{update};
-	my $setopt = $self->{config}->{set_option};
-	my $updopt = $self->{update}->{set_option} || [];
-	my %option = map { @$_ } @$setopt, @$updopt;
+	my $entry = ( scalar(@arg) > 1 ) ? [@arg] : $arg[0];
+	my $opt	  = "${name}:";
+	$self->{test_ctx} = Net::DNS::Resolver::Unbound::Context->new() unless $self->{update};
+	$self->{test_ctx}->set_option( $opt, @arg ) if defined $entry;	  # error check only
+	my $updopt = $self->{update}->{set_option};
+	my %option = map {@$_} @$updopt;
 
-	my $opt = "${name}:";
-	return $option{$opt} unless @arg;
-
-	my $arg = shift @arg;
-	delete $option{$opt};
-	if ( defined $arg ) {
-		$self->{ub_ctx}->set_option( $opt, $arg );
-		$option{$opt} = $arg;
+	unless (@arg) {
+		my $setopt = $self->{config}->{set_option};
+		my %config = map {@$_} @$setopt, @$updopt;
+		my $value  = $config{$opt};
+		return ref($value) ? @$value : $value;
 	}
 
+	my $state = $option{$opt};
+	if ( defined $state ) {					# second and subsequent entries
+		$option{$opt} = $state = [$state] unless ref $state;
+		push @$state, $entry;
+	} else {						# initial entry
+		$option{$opt} = ref($entry) ? [$entry] : $entry;
+	}
+
+	$option{$opt} = undef unless defined $entry;		# delete entire option
 	my @option = map { [$_, $option{$_}] } keys %option;
 	$self->{update}->{set_option} = \@option;
 	return;
@@ -530,10 +552,13 @@ sub _option {
 
 
 my %IP_conf = (
-	force_v4  => ['do-ip6'	   => 'no'],
-	force_v6  => ['do-ip4'	   => 'no'],
-	prefer_v4 => ['prefer-ip4' => 'yes'],
-	prefer_v6 => ['prefer-ip6' => 'yes'] );
+	force_v4  => ['do-ip6:'	    => 'no'],
+	force_v6  => ['do-ip4:'	    => 'no'],
+	prefer_v4 => ['prefer-ip4:' => 'yes'],
+	prefer_v6 => ['prefer-ip6:' => 'yes'] );
+my @IPpref = map {@$_} values %IP_conf;
+delete @IP_conf{qw(prefer_v4 prefer_v6)} if UNBOUND_VERSION < 11100;
+my @IPconf = sort keys %IP_conf;
 
 sub _finalise_config {
 	my $self = shift;
@@ -542,34 +567,38 @@ sub _finalise_config {
 	return unless $update;
 	my $ctx = $self->{ub_ctx} = Net::DNS::Resolver::Unbound::Context->new();
 
-	my $config = $self->{config} || { set_option => [] };	# merge config updates
-	my %config = ( %$config, %$update );
-
-	my $optref = delete $config{set_option};		# create option hash table
-	my %option = map { @$_ } @$optref;
-	my @prefer = keys %IP_conf;
-	delete @option{@prefer};
-	my @option = map { [$_, $option{$_}] } keys %option;	# reassemble set_option list
+	my $config = $self->{config};
+	my $cfgopt = delete $config->{set_option};		# extract option lists
+	my $updopt = delete $update->{set_option};
+	my %config = ( %$config, %$update );			# merge config updates
 
 	foreach my $name ( keys %config ) {			# rebuild unbound context
-		my $value = $config{$name};
-		if ( ref $value ) {
-			foreach my $arg (@$value) {
-				my @arg = map { ref($_) ? @$_ : $_ } $arg;
-				$ctx->$name(@arg);
-			}
-		} else {
-			$ctx->$name($value);
+		my @element = map { ref($_) ? @$_ : $_ } $config{$name};
+		foreach my $element (@element) {
+			my @arg = map { ref($_) ? @$_ : $_ } $element;
+			$ctx->$name(@arg);
 		}
 	}
 
-	foreach my $key (@prefer) {				# append IP preference
-		eval {
-			my ( $name, $arg ) = @{$IP_conf{$key}};
-			my $opt = "${name}:";
-			$ctx->set_option($opt,$arg);	# not implemented?
-			push @option, [$opt, $arg];
-		} if $self->$key;
+	my %option = map {@$_} @$cfgopt, @$updopt;		# merge options
+	delete @option{@IPpref};				# expunge IP preferences
+	my @option;
+
+	foreach my $name ( keys %option ) {			# set unbound options
+		my $value = $option{$name} || next;
+		push @option, [$name, $value];
+		my @element = map { ref($_) ? @$_ : $_ } $value;
+		foreach my $element (@element) {
+			my @arg = map { ref($_) ? @$_ : $_ } $element;
+			$ctx->set_option( $name, @arg );
+		}
+	}
+
+	foreach my $key ( grep { $self->$_ } @IPconf ) {	# append IP preference
+		my $arg = $IP_conf{$key};
+		$ctx->set_option(@$arg);
+		push @option, $arg;
+		last;
 	}
 
 	$config{set_option} = \@option if @option;
